/*
Copyright (c) 1998 - 2007, The Regents of the University of California
Produced at the Lawrence Livermore National Laboratory
All rights reserved.

This file is part of PyCXX. For details,see http://cxx.sourceforge.net/. The
full copyright notice is contained in the file COPYRIGHT located at the root
of the PyCXX distribution.

Redistribution  and  use  in  source  and  binary  forms,  with  or  without
modification, are permitted provided that the following conditions are met:

  - Redistributions of  source code must  retain the above  copyright notice,
    this list of conditions and the disclaimer below.
  - Redistributions in binary form must reproduce the above copyright notice,
    this  list of  conditions  and  the  disclaimer (as noted below)  in  the
    documentation and/or materials provided with the distribution.
  - Neither the name of the UC/LLNL nor  the names of its contributors may be
    used to  endorse or  promote products derived from  this software without
    specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT  HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR  IMPLIED WARRANTIES, INCLUDING,  BUT NOT  LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND  FITNESS FOR A PARTICULAR  PURPOSE
ARE  DISCLAIMED.  IN  NO  EVENT  SHALL  THE  REGENTS  OF  THE  UNIVERSITY OF
CALIFORNIA, THE U.S.  DEPARTMENT  OF  ENERGY OR CONTRIBUTORS BE  LIABLE  FOR
ANY  DIRECT,  INDIRECT,  INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT  LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR
SERVICES; LOSS OF  USE, DATA, OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER
CAUSED  AND  ON  ANY  THEORY  OF  LIABILITY,  WHETHER  IN  CONTRACT,  STRICT
LIABILITY, OR TORT  (INCLUDING NEGLIGENCE OR OTHERWISE)  ARISING IN ANY  WAY
OUT OF THE  USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/

#ifndef __PyCXX_wrap_python_hxx__
#define __PyCXX_wrap_python_hxx__

/* On some platforms we have to include time.h to get select defined */
#if !defined(__WIN32__) && !defined(WIN32) && !defined(_WIN32) && !defined(_WIN64)
#include <sys/time.h>
#endif

/* Prevent warnings */
#if defined(_XOPEN_SOURCE)
  #undef _XOPEN_SOURCE
#endif
#if defined(_POSIX_C_SOURCE)
  #undef _POSIX_C_SOURCE
#endif

/* Prevent multiple conflicting definitions of swab from stdlib.h and unistd.h
See also http://bugs.kde.org/show_bug.cgi?id=156950 */
#if defined(__sun) || defined(sun)
  #if defined(_XPG4)
    #undef _XPG4
  #endif
#endif

/* pull in python definitions */
#include <Python.h>

#endif
